#define NO_PORTB_PINCHANGES // 8,9,10,11,12,13
#define NO_PORTC_PINCHANGES // A0,A1,A2,A3,A4,A5
#include <PinChangeInt.h>


#include <Arduino.h>
#include "CAN_S88_Gateway.h"      // interface CAN - S88
#include <SPI.h>                  //  SPI library
#include <mcp_can.h>              // MCP2515 library

#define RETRY_CONNECTION 10    // retry connection
#define NB_DETECTOR  4         // number of detector (4 max) of 15 sensors (60 max by Gateway)

// pins allocation

const byte dataIn = 0;       // pin input from another arduino
const byte dataOut = 1;       // pin output to another arduino or console
const byte clockS88 = 2;     // S88 clock on pin 2
const byte PSS88 = 3;        // S88 PS on pin 3
const byte CANPin = 4;      // interrupt from CAN on pin 4
const byte CS = 10;         //  chip select

// CAN parameters

const int baudrate = CAN_500KBPS;       // can throughput 500 Kb/s
const byte Id_Mode = 0;                 // 11 bits of ident
MCP_CAN can(CS);                        // CAN instance
volatile bool FlagReceive = false;      // can interrupt flag

// timing variable

volatile byte clockCounter = 0;     // 16 tops clock

// data buffer

volatile unsigned int data[NB_DETECTOR];   // shift buffer

// sensors

volatile unsigned int sensors[NB_DETECTOR];   // buffer from sensor

// message CAN information

unsigned char gatewayNumber;   // number of the gateway       
unsigned long Id;
unsigned char Len = 0;        // length of data, 3 octets by message
unsigned char Buf[8];         // buffer of data : byte1 : Gateway number, byte2 : Detector number, byte3: data high, byte4 : data low

// CAN interrupt routine

void MCP2515_ISR() {FlagReceive = true;}

// PS signal from S88 interrupt routine

void PS() {
  cli();
  clockCounter = 0;                                             // init clock counter
  for (int i = 0; i< NB_DETECTOR; i++) {data[i] = sensors[i];} // load the shift buffer                                         
  sei();
}

// clock signal from S88 interrupt routine

void clock() {
  cli();                                                                  
  digitalWrite(dataOut, bitRead(data[0],clockCounter));                 // shift bit to other Arduino or console 
  for (int i = 0; i <NB_DETECTOR - 1; i++) {
    bitWrite(data[i],clockCounter, bitRead(data[i+1],clockCounter));
    } // shift all the buffer
  bitWrite(data[NB_DETECTOR -1],clockCounter,digitalRead(dataIn));     // shift from another Arduino
    clockCounter = (clockCounter + 1) % 16;                            // update the clock counter bits 0 to 15
  sei();                                                                  
}

// setup

void Setup(unsigned char GatewayNumber) {
    gatewayNumber = GatewayNumber;
    pinMode(clockS88,INPUT_PULLUP);                             // S88 clock in input
    PCintPort::attachInterrupt(clockS88,&clock,RISING);         // interrupt S88 clock
    pinMode(PSS88,INPUT_PULLUP);                                // S88 PS in input
    PCintPort::attachInterrupt(PSS88,&PS,RISING);               // interrupt PS
    pinMode(dataOut,OUTPUT);                                    // data output
    digitalWrite(dataOut,LOW);                                  // led off
    pinMode(dataIn,INPUT_PULLUP);                               // data input
    int repeat = RETRY_CONNECTION;                              // try to open the CAN
    while (repeat > 0) {
        if (CAN_OK == can.begin(baudrate)){break;}
        else {repeat--;}
        delay(200);
    }
    pinMode(CANPin,INPUT);                                        // CAN int pin
    PCintPort::attachInterrupt(PCINT1,&MCP2515_ISR,FALLING);     // CAN interrupt
    for (int i = 0; i< NB_DETECTOR; i++) {data[i] = 0xFFFF; sensors[i] = 0x0000;}  // init the buffers
}

// loop

void Loop() {

  // translate receive message from CAN to sensor buffer
  
   if (FlagReceive) {FlagReceive = false;   // if receive message from CAN bus
       while (CAN_MSGAVAIL == can.checkReceive()) {
                can.readMsgBufID(&Id,&Len,Buf);
                if (Buf[0] == gatewayNumber) {sensors[Buf[1]-1] = (Buf[2] << 8) + Buf[3];}
            }
          }
}



